UC_PREFIX TEXTEQU <u>

UC  TEXTEQU <UCCSTR>
UCC TEXTEQU <UCCSTR>

;/********************************************************************/
;/*                     fnx - macro procedure                        */
;/* This macro enhanced the INVOKE-directive:                        */
;/*   - It adds support for quoted ASCII or unicode strings.         */
;/*     The strings can be either enclosed by double quotes or by    */
;/*     single quotation marks.                                      */
;/*     The kind of created string (Unicode or ASCII) depends on the */
;/*     __UNICODE__ equate. If this equte is defined and has a       */
;/*     nonzero value, a unicode string is created. However, creation*/
;/*     of Unicode strings can be forced by using the 'u'-prefix:    */
;/*                 u"my string" or u'my string'                     */
;/*     MASM's reserved characters like exclamation marks, angel     */
;/*     brackets and single brackets [,...] can not be used.         */
;/*   - ADDR-expressions can be shorten by using a ampersand '&':    */
;/*         fn MessageBoxW,0,&wsz[0],L'xyz',0                        */
;/*   - a optional destination can be specified in front of the      */
;/*     function:                                                    */
;/*         fn dest=FncName,...                                      */
;/*                                                                  */
;/* Example:                                                         */
;/*     fnx MessageBox,0,"my string",&wsz[0],0                       */
;/*     fnx hWnd = CreateWindowEx,...                                */
;/*                                                     qWord, 2011  */
;/********************************************************************/
fnx macro FncName:req,args:VARARG

    ;/* check if a optional destination is specified */
    fnex_flag = 0
    IF @InStr(1,<&FncName>,<=>)
        fnex_flag = 1
        fnex_dest SUBSTR <&FncName>,1,@InStr(1,<&FncName>,<=>)-1
        fnex_arg TEXTEQU <invoke >,@SubStr(<&FncName>,@InStr(1,<&FncName>,<=>)+1)
    ELSE    
        fnex_arg TEXTEQU <invoke FncName>
    ENDIF

    ;/* process argument list and append it to the invoke-call */
    fnex_arg CATSTR fnex_arg,repargs(&args)

    ;/* place the function call */
    fnex_arg

    ;/* if available, fill the specified destination */    
    IF fnex_flag EQ 1
        mov fnex_dest,eax
    ENDIF
endm

;/********************************************************************/
;/*                     rvx - macro function                         */
;/* This macro is the function-version of the fnx-macro.             */
;/*                                                                  */
;/* Example:                                                         */
;/*      mov edi,rv(myFunction,L"my string",&wsz[0],...)             */
;/*      .break .if !rv(dest=myFunction,...)                         */
;/*                                                     qWord, 2011  */
;/********************************************************************/
rvx macro FncName:req,args:VARARG
    
    ;/* check if a optional destination is specified */
    rvex_flag = 0
    IF @InStr(1,<&FncName>,<=>)
        rvex_flag = 1
        rvex_dest SUBSTR <&FncName>,1,@InStr(1,<&FncName>,<=>)-1
        rvex_arg TEXTEQU <invoke >,@SubStr(<&FncName>,@InStr(1,<&FncName>,<=>)+1)
    ELSE    
        rvex_arg TEXTEQU <invoke FncName>
    ENDIF

    ;/* process argument list and append it to the invoke-call */
    rvex_arg CATSTR rvex_arg,repargs(&args)
    
    ;/* place the function call */
    rvex_arg
    
    ;/* if available, fill the specified destination */
    IF rvex_flag EQ 1
        mov rvex_dest,eax
    ENDIF
    EXITM <eax>
endm

;/*******************************************************************/
;/*                     simplwsz , macro procedure                  */
;/* Creates a zero terminated Unicode string in current segment.    */                         
;/* Parameters:                                                     */
;/*     lbl : The label/name of the string                          */
;/*     txt : string, either enclosed by single quotation marks     */
;/*           or by double quotes. You cannot use MASM's            */
;/*           reserved characters like exclamation marks,           */
;/*           angel brackets, single brackets,... .                 */
;/*           <""> or <''> cause the cration of an empty string.    */
;/* Return: an error is indicated by setting the equate             */
;/*         smplwsz_error to 1                                      */
;/* Example:                                                        */
;/*     .data                                                       */
;/*         simplwsz myLabel,"unicode 123'4"                        */
;/*         IF smplwsz_error                                        */
;/*            ;; error                                             */
;/*         ELSE                                                    */
;/*            ;; OK, no error                                      */
;/*         ENDIF                                                   */
;/*     .code                                          qWord, 2011  */
;/*******************************************************************/
simplwsz macro lbl:req,txt:req

    smplwsz_error = 0

    ;/* remove leading blanks (trim) */ 
    FOR arg,<&txt>
        simplwsz_arg TEXTEQU <&arg>
        EXITM
    ENDM
    
    ;/* exit if argument is blank */
    IFB simplwsz_arg
        smplwsz_error = 1
        EXITM
    ENDIF
    
    simplwsz_pos1 INSTR 1,simplwsz_arg,<">
    simplwsz_pos2 INSTR 1,simplwsz_arg,<'>

    align 2
    lbl LABEL WORD

    ;/* check wether the string is enclosed by     */
    ;/* double quotes or by single quotation marks */
    IF simplwsz_pos1 EQ 1
        ;/* loop through all cahr. after removing the enclosing */ 
        ;/* double quotes                                       */
    %   FORC char,<@SubStr(%simplwsz_arg,2,@SizeStr(%simplwsz_arg)-2)>
            dw "&char"
        ENDM
    ELSEIF simplwsz_pos2 EQ 1
    %   FORC char,<@SubStr(%simplwsz_arg,2,@SizeStr(%simplwsz_arg)-2)>
            dw '&char'
        ENDM
    ELSE
        ;/* invalid argument *
        smplwsz_error = 1
    ENDIF
    ;/* term. zero */
    dw 0 
    EXITM
endm

;/*****************************************************************/ 
;/*                 repargs , macro function                      */
;/* Parse the arguments list 'args' and replace:                  */
;/*     - String literals by the corresponding OFFSET-expression  */
;/*       after creating an anonym string in the .data-section    */
;/*     - leading ampersands (&) by the ADDR operator             */
;/* All other arguments are left untouched.                       */
;/*                                                               */
;/* Details:                                                      */
;/*     This macro creates, according to the __UNICODE__-equate,  */
;/*     either ASCII or Unicode strings. However, Unicode strings */
;/*     can be forced by using the 'u'-Prefix:                    */
;/*       u"my unicode string"  or u'my "quoted" string'          */
;/*     Furthermore, ampersands are replaced by <ADDR >, if they  */
;/*     are the first character of the argument:                  */
;/*        ..., &sz[0],...    ->   ..., ADDR sz[0],...            */
;/*                                                               */
;/* Return:                                                       */
;/*     The processed argument list. If a nonempty list is passed */
;/*     to the macro, the returned list is always leaded          */
;/*     by a comma:                                               */
;/*         <arg1,arg2,...> -> <,newArg1,newArg2,...>             */
;/*     The number of arguments in list is stored in              */
;/*     the equate repargs_cntr.                                  */
;/*                                                               */
;/* Remarks:                                                      */
;/*     This macro is designed to process arguments lists for     */
;/*     the INVOKE-directive.                                     */
;/*     The macro assumes to be called from the .code-section     */
;/*                                                               */
;/* Example:                                                      */
;/*     invk_txt TEXTEQU <invoke FncName>                         */
;/*     invk_txt CATSTR invk_txt,repargs(ArgumentList)            */
;/*     invk_txt                                                  */
;/*                                                               */
;/* For advanced users: The 'u'-prefix was chosen, because a bug  */
;/* in ML v6.14.8444 does not allow to use the well known         */
;/* L-prefix (c/c++). This bug is not present in newer version,   */
;/* thus the L-prefix can be used by declaring text macro         */
;/* UC_PREFIX = <L>                                               */
;/*                                                  qWord, 2011  */
;/*****************************************************************/ 
repargs macro args:VARARG
    
    ;/* initialize global counter which */
    ;/* is used for label-creation.      */
    IFNDEF repargs_glb_cntr
        repargs_glb_cntr = 0
    ENDIF

    repargs_unicode = 0
    IFDEF __UNICODE__
        IF __UNICODE__
            repargs_unicode = 1
        ENDIF
    ENDIF

    repargs_cntr = 0
    repargs_args TEXTEQU <>
%   FOR arg,<&args>
        repargs_txt  TEXTEQU <&arg>
        repargs_prefix1 CATSTR UC_PREFIX,<">
        repargs_prefix2 CATSTR UC_PREFIX,<'>
       
        ;/* analyze current argument */
        repargs_pos1 INSTR 1,<&arg>,<">
        repargs_pos2 INSTR 1,<&arg>,<'>
        repargs_pos3 INSTR 1,<&arg>,repargs_prefix1
        repargs_pos4 INSTR 1,<&arg>,repargs_prefix2
        repargs_pos5 INSTR 1,<&arg>,<!&>

        IF repargs_pos1 EQ 1 OR repargs_pos2 EQ 1
            .data
                ;/* creates strings according to the UNICODE-equate */
                IFE repargs_unicode
                    ;/ create a ASCII string */
                    IF @SizeStr(<%repargs_txt>) GT 2
                        @CatStr(<$$_szlbl_>,%repargs_glb_cntr) db repargs_txt,0
                    ELSE
                        @CatStr(<$$_szlbl_>,%repargs_glb_cntr) db 0
                    ENDIF               
                    repargs_args TEXTEQU repargs_args,<,OFFSET $$_szlbl_>,%repargs_glb_cntr
                ELSE
                    ;/ create a Unicode string */
                    IF @SizeStr(<%repargs_txt>) GT 2
                        simplwsz @CatStr(<$$_wszlbl_>,%repargs_glb_cntr),<%repargs_txt>
                    ELSE
                        @CatStr(<$$_wszlbl_>,%repargs_glb_cntr) dw 0
                    ENDIF
                    repargs_args TEXTEQU repargs_args,<,OFFSET $$_wszlbl_>,%repargs_glb_cntr
                ENDIF
            .code
            repargs_glb_cntr = repargs_glb_cntr + 1
        ELSEIF repargs_pos3 EQ 1 OR repargs_pos4 EQ 1
            ;/* remove 'u'-prefix  */
            repargs_ps SIZESTR UC_PREFIX
            repargs_txt SUBSTR repargs_txt,repargs_ps+1
            
            ;/* create Unicode string */
            .data
                IF @SizeStr(<%repargs_txt>) GT 2
                    simplwsz @CatStr(<$$_wszlbl_>,%repargs_glb_cntr),<%repargs_txt>
                ELSE
                    @CatStr(<$$_wszlbl_>,%repargs_glb_cntr) dw 0
                ENDIF
            .code
            repargs_args TEXTEQU repargs_args,<,OFFSET $$_wszlbl_>,%repargs_glb_cntr
            repargs_glb_cntr = repargs_glb_cntr + 1
        ELSEIF repargs_pos5 EQ 1
            ;/* replace '&' with <ADDR > and append argument to list */
            repargs_args TEXTEQU repargs_args,<,>,<ADDR >,@SubStr(%repargs_txt,2,)
        ELSE
            ;/* record unprocessed argument */
            repargs_args TEXTEQU repargs_args,<,>,repargs_txt
        ENDIF
        repargs_cntr = repargs_cntr + 1
    ENDM
    EXITM repargs_args
endm


;/*****************************************************************/
;/*             UCSTR - macro procedure                           */
;/* This macro creates a Unicode string in the current segment    */
;/* Parameters:                                                   */
;/*    lbl:    [optional] name of string                          */
;/*    args:   one ore more string literals, either enclosed by   */
;/*            single quotation marks or by double quotes.        */
;/*            Constants can also be used, blank arguments        */
;/*            force an error. The maximal number of characters   */
;/*            is something about 240.                            */
;/*                                                               */
;/* Remarks:   A termination zero must add by user!               */
;/*            Named strings can be used with the SIZEOF and      */
;/*            LENGTHOF operator. This macro wont work with       */
;/*            angle brackets and exclemation marks.              */
;/* Example:                                                      */
;/*          UCSTR myStr, "my string",13,10,'"quoted"',0          */
;/*                                                  qWord, 2011  */
;/*****************************************************************/
UCSTR macro lbl,args:VARARG
    
    ;/* initialize counter used for label creation */
    IFNDEF ucstr_lbl_cntr
        ucstr_lbl_cntr = 0
    ENDIF
    
    ;/* if required, create a label */
    IFB <&lbl>
        ucstr_lbl TEXTEQU <$$_WSTR_>,%ucstr_lbl_cntr
        ucstr_lbl_cntr = ucstr_lbl_cntr + 1
    ELSE
        ucstr_lbl TEXTEQU <&lbl>
    ENDIF

    ucstr_size = 0
    ucstr_flg = 0  ; 0 => invalid argument
                   ; 1 => double quotes are used
                   ; 2 => single quotation marks are used
                   ; 3 => constant 
    ucstr_iarg = 1
    
    ;/* The following loop count the number of required WCHAR's */
%   FOR _arg,<&args>
        ;/* loop through all characters */
        ucstr_flg = 0
        FORC char,<&_arg>
            IF ucstr_flg NE 0
                ucstr_pos INSTR 1,<"'>,<&char>
                ;/* end of quoted string reached? */
                IF ucstr_pos EQ ucstr_flg
                    EXITM
                ENDIF
                ucstr_size = ucstr_size + 1
            ELSE
                ;/* This part is only one time executet for each argument _arg. */
                ;/* It determine wheter a string litreal, a numeric constant or */
                ;/* an invalid arguemten is present.                            */
                ucstr_flg INSTR 1,<"'>,<&char>
                IFE ucstr_flg
                    ;/* constant value ? */
                    IF (OPATTR _arg) AND 100y
                        ucstr_flg = 3
                    ENDIF
                    EXITM
                ENDIF
            ENDIF
        ENDM
        IF ucstr_flg EQ 3
            ;/* numeric constant */
            ucstr_size = ucstr_size + 1
        ELSEIFE ucstr_flg
            ;/* invalid argument detected -> exit */
            EXITM
        ENDIF
        ucstr_iarg = ucstr_iarg + 1     
    ENDM
    IFE ucstr_flg
    %   .err <invalid string specifier : argument : @CatStr(%ucstr_iarg)>
        EXITM
    ENDIF
    
    ;/* allocate string in current segment */
    align 2
    ucstr_lbl WORD ucstr_size dup (?)
    org $-ucstr_size*2
    
    ;/* This loop is nearly identically to the one above, excpet that char. are emitted */
%   FOR _arg,<&args>
        ucstr_flg = 0
        FORC char,<&_arg>
            IF ucstr_flg NE 0
                ucstr_pos INSTR 1,<"'>,<&char>
                IF ucstr_pos EQ ucstr_flg
                    EXITM
                ELSE
                    ucstr_char CATSTR <dw >,ucstr_quote,<&char>,ucstr_quote
                    ucstr_char
                ENDIF
            ELSE
                ucstr_flg INSTR 1,<"'>,<&char>
                IFE ucstr_flg
                    IF (OPATTR _arg) AND 100y
                        ucstr_flg = 3
                    ENDIF
                    EXITM
                ENDIF
                ucstr_quote TEXTEQU <&char>
            ENDIF
        ENDM
        IF ucstr_flg EQ 3
            dw _arg
        ELSEIFE ucstr_flg
            EXITM
        ENDIF
        ucstr_iarg = ucstr_iarg + 1     
    ENDM

endm

;/***************************************************/
;/*               uc$ , macro function              */
;/* This macro is the function-version of UCSTR.    */
;/* In contrast to UCSTR, the created string is     */
;/* zero terminated. The macro assumes to be called */
;/* from the .code-section and places the string    */
;/* in the .data-section.                           */
;/* Example:                                        */
;/*      mov esi,uc$("my string",13,10,'xxyz')      */
;/*                                    qWord, 2011  */
;/***************************************************/
uc$ macro args:VARARG
	.data
    	UCSTR ,args,0
    .code
    ucsz_retval TEXTEQU <OFFSET >,ucstr_lbl
    EXITM ucsz_retval
endm

;/*****************************************************************/
;/*             UCCSTR - macro procedure                          */
;/* This macro creates a Unicode string in the current segment    */
;/* This macro is identically to the UCSTR-macro, except, that it */
;/* adds support for some escape sequences:                       */
;/*     \\  ->  "\"                                               */
;/*     \t  ->  tab                                               */
;/*     \n  ->  new line (13,10)                                  */
;/*     \x  ->  "!"                                               */
;/*     \a  ->  "("                                               */
;/*     \b  ->  ")"                                               */
;/*     \l  ->  "<"                                               */
;/*     \r  ->  ">"                                               */
;/*     \q  ->  double quote '"'                                  */
;/*                                                               */
;/* Example:                                                      */
;/*       UCCSTR myStr, "line 1\nline2\nI'm here\x",0             */
;/*                                                  qWord, 2011  */
;/*****************************************************************/
UCCSTR macro lbl,args:VARARG

    IFNDEF uccstr_lbl_cntr
        uccstr_lbl_cntr = 0
    ENDIF
    IFB <&lbl>
        uccstr_lbl TEXTEQU <anonym_WSTR_>,%uccstr_lbl_cntr
        uccstr_lbl_cntr = uccstr_lbl_cntr + 1
    ELSE
        uccstr_lbl TEXTEQU <&lbl>
    ENDIF

    uccstr_size = 0
    uccstr_flg = 0
    uccstr_iarg = 1
    uccstr_esc = 0
    FOR _arg,<&args>
        uccstr_flg = 0
        FORC char,<&_arg>
            IF uccstr_flg NE 0
                uccstr_pos INSTR 1,<"'\>,<&char>
                IF uccstr_pos EQ uccstr_flg
                    EXITM
                ELSEIF uccstr_pos EQ 3 AND uccstr_esc EQ 0
                uccstr_esc = 1
                ELSE
                    IF uccstr_esc
                        uccstr_pos INSTR 1,<\ablrxqtn>,<&char>
                        
                        IFE uccstr_pos
                            uccstr_flg=0
                            EXITM
                        ENDIF
                        uccstr_size = uccstr_size + uccstr_pos/9
                        uccstr_esc = 0
                    ENDIF
                    uccstr_size = uccstr_size + 1
                ENDIF
            ELSE
                uccstr_flg INSTR 1,<"'>,<&char>
                IFE uccstr_flg
                    IF (OPATTR _arg) AND 100y
                        uccstr_flg = 3
                    ENDIF
                    EXITM
                ENDIF
            ENDIF
        ENDM
        IF uccstr_flg EQ 0 OR uccstr_esc NE 0
            EXITM
        ELSEIF uccstr_flg EQ 3
            uccstr_size = uccstr_size + 1
        ENDIF       
        uccstr_iarg = uccstr_iarg + 1
    ENDM
    IF uccstr_flg EQ 0 OR uccstr_esc NE 0
        IF uccstr_esc
        %   .err <invalid escape sequence : argument : @CatStr(%uccstr_iarg)>
        ELSE
        %   .err <invalid string specifier : argument : @CatStr(%uccstr_iarg)>
        ENDIF
        EXITM
    ENDIF
    
    align 2
    uccstr_lbl WORD uccstr_size dup (?)
    org $-uccstr_size*2
    
    uccstr_esc = 0
    FOR _arg,<&args>
        uccstr_flg = 0
        FORC char,<&_arg>
            IF uccstr_flg NE 0
                uccstr_pos INSTR 1,<"'\>,<&char>
                IF uccstr_pos EQ uccstr_flg
                    EXITM
                ELSEIF uccstr_pos EQ 3 AND uccstr_esc EQ 0
                    uccstr_esc = 1
                ELSE
                    IFE uccstr_esc
                        uccstr_char CATSTR <dw >,uccstr_quote,<&char>,uccstr_quote
                        uccstr_char
                    ELSE
                        uccstr_pos INSTR 1,<\ablrxqtn>,<&char>
                        IFE uccstr_pos
                            uccstr_flg=0
                            EXITM
                        ENDIF
                        uccstr_char SUBSTR <  5ch28h29h3ch3eh21h22h09h0ah,0dh>,uccstr_pos*3,3+4*(uccstr_pos/9)
                        uccstr_esc = 0
                        dw uccstr_char
                    ENDIF
                ENDIF
            ELSE
                uccstr_flg INSTR 1,<"'>,<&char>
                IFE uccstr_flg
                    IF (OPATTR _arg) AND 100y
                        uccstr_flg = 3
                    ENDIF
                    EXITM
                ENDIF
                uccstr_quote TEXTEQU <&char>
            ENDIF
        ENDM
        IF uccstr_flg EQ 3
            dw _arg
        ENDIF       
    ENDM

endm

;/***************************************************/
;/*               ucc$ , macro function             */
;/* This macro is the function-version of UCCSTR.   */
;/* In contrast to UCCSTR, the created string is    */
;/* zero terminated. The macro assumes to be called */
;/* from the .code-section and places the string    */
;/* in the .data-section.                           */
;/*     mov esi,ucc$("\lHello World\r\n:-\b")       */
;/*                                    qWord, 2011  */
;/***************************************************/
ucc$ macro args:VARARG
    .data
    	UCCSTR ,args,0
    .code
    uccsz_retval TEXTEQU <OFFSET >,uccstr_lbl
    EXITM uccsz_retval
endm
